(******************************************************************************

______________________________________________________________________________

ColBuild v0.10                                               (c) 2009-12 Pepak
http://www.pepak.net/e-books/prs-505-sbirky/              http://www.pepak.net
______________________________________________________________________________


Copyright (c) 2009-12 Pepak (http://www.pepak.net)
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of Pepak nor the
      names of his contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL PEPAK BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

******************************************************************************)

unit uCalibreEbooks;

interface

uses
  SysUtils, Classes, Windows, uEbooks, uSystem;

type
  TCalibreEbookFile = class(TEbookFile)
  private
  protected
    function CalibreEbookMetaExe: string;
    procedure LoadMetaData; override;
  public
    class function IsSupportedFile(const AFileName: string): boolean; override;
    constructor Create(const AFileName: string); override;
  end;

var
  CalibrePath: string = '';

implementation

{ TCalibreEbookFile }

class function TCalibreEbookFile.IsSupportedFile(const AFileName: string): boolean;
const
  Supported: array[0..12] of string = ('azw', 'azw1', 'azw4', 'epub', 'htmlz', 'lrf', 'mobi', 'pdb', 'pdf', 'prc', 'rtf', 'tpz', 'txtz');
begin
  Result := IsSupportedFileByExtension(AFileName, Supported);
end;

constructor TCalibreEbookFile.Create(const AFileName: string);
begin
  inherited;
end;

function TCalibreEbookFile.CalibreEbookMetaExe: string;
const
  EBOOK_META = 'ebook-meta.exe';
begin
  if CalibrePath = '' then
    Result := EBOOK_META
  else
    Result := AnsiQuotedStr(IncludeTrailingPathDelimiter(CalibrePath) + EBOOK_META, '"');
end;

procedure TCalibreEbookFile.LoadMetaData;
var
  Output: TStringList;
  ResCode: DWORD;
  i, j: Integer;
  s, Name, Value: string;
begin
  inherited LoadMetaData;
  Output := TStringList.Create;
  try
    if GetConsoleOutput(CalibreEbookMetaExe + ' ' + AnsiQuotedStr(FileName, '"'), TStringList(nil), Output, TStringList(nil), ResCode, SW_HIDE) then
      if ResCode = 0 then
        for i := 0 to Pred(Output.Count) do
          begin
          s := Output[i];
          j := Pos(':', s);
          if j > 0 then
            begin
            Name := Trim(Copy(s, 1, Pred(j)));
            Value := Trim(Copy(s, Succ(j), MaxInt));
            if AnsiCompareText(Name, 'title') = 0 then
              Title := Value
            else if AnsiCompareText(Name, 'title sort') = 0 then
              TitleSort := Value
            else if AnsiCompareText(Name, 'author(s)') = 0 then
              begin
              j := Pos('[', Value);
              if j > 0 then
                begin
                Author := Trim(Copy(Value, 1, Pred(j)));
                AuthorSort := Trim(Copy(Value, Succ(j), MaxInt));
                j := Pos(']', AuthorSort);
                if j > 0 then
                  AuthorSort := Trim(Copy(AuthorSort, 1, Pred(j)));
                end
              else
                begin
                Author := Value;
                AuthorSort := '';
                end;
              end
            else if AnsiCompareText(Name, 'series') = 0 then
              begin
              j := Pos('#', Value);
              if j > 0 then
                begin
                Series := Trim(Copy(Value, 1, Pred(j)));
                SeriesIndex := StrToIntDef(Trim(Copy(Value, Succ(j), MaxInt)), -1);
                end
              else
                begin
                Series := Value;
                SeriesIndex := -1;
                end;
              end
            else if AnsiCompareText(Name, 'languages') = 0 then
              Language := Value;
            end;
          end;
  finally
    FreeAndNil(Output);
    end;
end;

initialization
  TEbookFileList.RegisterEbookClass(TCalibreEbookFile, 'ZZZ');

end.
