(******************************************************************************

______________________________________________________________________________

ColBuild v0.10                                               (c) 2009-12 Pepak
http://www.pepak.net/e-books/prs-505-sbirky/              http://www.pepak.net
______________________________________________________________________________


Copyright (c) 2009-12 Pepak (http://www.pepak.net)
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of Pepak nor the
      names of his contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL PEPAK BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

******************************************************************************)

unit uEbooks;

interface

uses
  SysUtils, Classes, Windows, Graphics, uSystem;

type
  TEbookFileClass = class of TEbookFile;
  TEbookFile = class
    private
      fFileName: string;
      fSize: int64;
      fCreationDate: TDateTime;
      fModificationDate: TDateTime;
      fAccessDate: TDateTime;
      fMetadataLoaded: boolean;
      fTitle: string;
      fTitleSort: string;
      fAuthor: string;
      fAuthorSort: string;
      fSeries: string;
      fSeriesIndex: integer;
      fLanguage: string;
      fCover: TGraphic;
      function GetAccessDate: TDateTime;
      function GetAuthor: string;
      function GetAuthorSort: string;
      function GetCover: TGraphic;
      function GetCreationDate: TDateTime;
      function GetLanguage: string;
      function GetModificationDate: TDateTime;
      function GetSeries: string;
      function GetSeriesIndex: integer;
      function GetSize: int64;
      function GetTitle: string;
      function GetTitleSort: string;
      procedure SetAuthor(const Value: string);
      procedure SetAuthorSort(const Value: string);
      procedure SetLanguage(const Value: string);
      procedure SetSeries(const Value: string);
      procedure SetSeriesIndex(const Value: integer);
      procedure SetTitle(const Value: string);
      procedure SetTitleSort(const Value: string);
    protected
      class function IsSupportedFileByExtension(const AFileName: string; const SupportedExtensions: array of string): boolean;
      procedure ClearMetadata; virtual;
      procedure LoadMetaData; virtual;
      procedure SetCover(const Value: TGraphic);
    public
      class function IsSupportedFile(const AFileName: string): boolean; virtual; abstract;
      constructor Create(const AFileName: string); virtual;
      destructor Destroy; override;
      procedure Refresh;
      procedure Assign(Source: TEbookFile);
      procedure GenerateCover; virtual;
      property FileName: string read fFileName;
      property Size: int64 read GetSize;
      property CreationDate: TDateTime read GetCreationDate;
      property ModificationDate: TDateTime read GetModificationDate;
      property AccessDate: TDateTime read GetAccessDate;
      property Title: string read GetTitle write SetTitle;
      property TitleSort: string read GetTitleSort write SetTitleSort;
      property Author: string read GetAuthor write SetAuthor;
      property AuthorSort: string read GetAuthorSort write SetAuthorSort;
      property Series: string read GetSeries write SetSeries;
      property SeriesIndex: integer read GetSeriesIndex write SetSeriesIndex;
      property Language: string read GetLanguage write SetLanguage;
      property Cover: TGraphic read GetCover;
    end;

  TEbookFileList = class
    private
      fList: TList;
      fOwnsObjects: boolean;
      fSkipMetadataExtraction: boolean;
      fGenerateDummyCovers: boolean;
      function GetItem(Index: integer): TEbookFile;
      procedure SetItem(Index: integer; const Value: TEbookFile);
    protected
      class function EbookFileFromFileName(const AFileName: string; out AEbookFile: TEbookFile): boolean;
      property List: TList read fList;
    public
      class procedure RegisterEbookClass(AClass: TEbookFileClass; const APriority: string = '');
      constructor Create(AOwnsObjects: boolean = False);
      destructor Destroy; override;
      procedure Clear;
      property OwnsObjects: boolean read fOwnsObjects write fOwnsObjects;
      property Items[Index: integer]: TEbookFile read GetItem write SetItem; default;
      function Count: integer;
      function Add(const Value: TEbookFile): integer;
      procedure Insert(const Index: integer; const Value: TEbookFile);
      procedure Delete(const Index: integer);
      function IndexOf(const Value: TEbookFile): integer;
      procedure Sort;
      property SkipMetadataExtraction: boolean read fSkipMetadataExtraction write fSkipMetadataExtraction;
      property GenerateDummyCovers: boolean read fGenerateDummyCovers write fGenerateDummyCovers;
      function AddFile(const AFileName: string): integer;
    end;

implementation

{ TEbookFileList }

var
  RegisteredEbookClasses: TStringList;

class procedure TEbookFileList.RegisterEbookClass(AClass: TEbookFileClass; const APriority: string);
begin
  if AClass <> nil then
    RegisteredEbookClasses.AddObject(APriority, TObject(AClass));
end;

class function TEbookFileList.EbookFileFromFileName(const AFileName: string; out AEbookFile: TEbookFile): boolean;
var
  i: Integer;
begin
  Result := False;
  AEbookFile := nil;
  for i := 0 to Pred(RegisteredEbookClasses.Count) do
    if TEbookFileClass(RegisteredEbookClasses.Objects[i]).IsSupportedFile(AFileName) then
      begin
      AEbookFile := TEbookFileClass(RegisteredEbookClasses.Objects[i]).Create(AFileName);
      Result := True;
      Break;
      end;
end;

constructor TEbookFileList.Create(AOwnsObjects: boolean);
begin
  inherited Create;
  fList := TList.Create;
  fOwnsObjects := AOwnsObjects;
end;

destructor TEbookFileList.Destroy;
begin
  Clear;
  FreeAndNil(fList);
  inherited;
end;

function TEbookFileList.GetItem(Index: integer): TEbookFile;
begin
  Result := TEbookFile(List[Index]);
end;

procedure TEbookFileList.SetItem(Index: integer; const Value: TEbookFile);
begin
  if OwnsObjects then
    if Items[Index] <> Value then
      Items[Index].Free;
  List[Index] := Value;
end;

function SortByNameFunction(Item1, Item2: Pointer): integer;
begin
  Result := AnsiCompareStr(TEbookFile(Item1).FileName, TEbookFile(Item2).FileName);
end;

procedure TEbookFileList.Sort;
begin
  List.Sort(SortByNameFunction);
end;

function TEbookFileList.Count: integer;
begin
  Result := List.Count;
end;

procedure TEbookFileList.Clear;
var
  i: integer;
begin
  if OwnsObjects then
    for i := 0 to Pred(Count) do
      TObject(List[i]).Free;
  List.Clear;
end;

function TEbookFileList.Add(const Value: TEbookFile): integer;
begin
  Result := List.Add(Value);
end;

function TEbookFileList.IndexOf(const Value: TEbookFile): integer;
begin
  Result := List.IndexOf(Value);
end;

procedure TEbookFileList.Insert(const Index: integer; const Value: TEbookFile);
begin
  List.Insert(Index, Value);
end;

procedure TEbookFileList.Delete(const Index: integer);
begin
  List.Delete(Index);
end;

function TEbookFileList.AddFile(const AFileName: string): integer;
var
  EbookFile: TEbookFile;
begin
  if EbookFileFromFileName(AFileName, EbookFile) then
    begin
    if not SkipMetadataExtraction then
      begin
      EbookFile.Refresh;
      if EbookFile.Cover = nil then
        if GenerateDummyCovers then
          EbookFile.GenerateCover;
      end;
    Result := Add(EbookFile);
    end
  else
    Result := -1;
end;

{ TEbookFile }

class function TEbookFile.IsSupportedFileByExtension(const AFileName: string; const SupportedExtensions: array of string): boolean;
var
  Ext: string;
  i: integer;
begin
  Result := False;
  Ext := Copy(ExtractFileExt(AFileName), 2, MaxInt);
  for i := 0 to Pred(Length(SupportedExtensions)) do
    if AnsiCompareText(Ext, SupportedExtensions[i]) = 0 then
      begin
      Result := True;
      Break;
      end;
end;

constructor TEbookFile.Create(const AFileName: string);
begin
  inherited Create;
  fFileName := AFileName;
  fMetadataLoaded := False;
end;

destructor TEbookFile.Destroy;
begin
  ClearMetadata;
  inherited;
end;

procedure TEbookFile.GenerateCover;
const
  DUMMY_WIDTH = 80;
  DUMMY_HEIGHT = 80;
  DUMMY_TEXT = '?';
var
  WorkBmp, Bmp: TBitmap;
  W, H: integer;
begin
  WorkBmp := TBitmap.Create;
  try
    WorkBmp.Canvas.Font.Color := clRed;
    WorkBmp.Canvas.Font.Style := [fsBold];
    WorkBmp.Canvas.Font.Size := 48;
    W := WorkBmp.Canvas.TextWidth(DUMMY_TEXT);
    H := WorkBmp.Canvas.TextHeight(DUMMY_TEXT);
    WorkBmp.SetSize(W + 4, H + 4);
    //WorkBmp.Canvas.Rectangle(0, 0, Pred(W), Pred(H));
    WorkBmp.Canvas.TextOut(2, 2, DUMMY_TEXT);
    Bmp := TBitmap.Create;
    try
      Bmp.SetSize(DUMMY_WIDTH, DUMMY_HEIGHT);
      Bmp.Canvas.StretchDraw(Rect(0, 0, DUMMY_WIDTH-1, DUMMY_HEIGHT-1), WorkBmp);
      SetCover(Bmp);
    except
      FreeAndNil(Bmp);
      Raise;
      end;
  finally
    FreeAndNil(WorkBmp);
    end;
end;

function TEbookFile.GetAccessDate: TDateTime;
begin
  if not fMetadataLoaded then
    LoadMetadata;
  Result := fAccessDate;
end;

function TEbookFile.GetAuthor: string;
begin
  if not fMetadataLoaded then
    LoadMetadata;
  Result := fAuthor;
end;

function TEbookFile.GetAuthorSort: string;
begin
  if not fMetadataLoaded then
    LoadMetadata;
  Result := fAuthorSort;
end;

function TEbookFile.GetCover: TGraphic;
begin
  if not fMetadataLoaded then
    LoadMetadata;
  Result := fCover;
end;

function TEbookFile.GetCreationDate: TDateTime;
begin
  if not fMetadataLoaded then
    LoadMetadata;
  Result := fCreationDate;
end;

function TEbookFile.GetLanguage: string;
begin
  if not fMetadataLoaded then
    LoadMetadata;
  Result := fLanguage;
end;

function TEbookFile.GetModificationDate: TDateTime;
begin
  if not fMetadataLoaded then
    LoadMetadata;
  Result := fModificationDate;
end;

function TEbookFile.GetSeries: string;
begin
  if not fMetadataLoaded then
    LoadMetadata;
  Result := fSeries;
end;

function TEbookFile.GetSeriesIndex: integer;
begin
  if not fMetadataLoaded then
    LoadMetadata;
  Result := fSeriesIndex;
end;

function TEbookFile.GetSize: int64;
begin
  if not fMetadataLoaded then
    LoadMetadata;
  Result := fSize;
end;

function TEbookFile.GetTitle: string;
begin
  if not fMetadataLoaded then
    LoadMetadata;
  Result := fTitle;
end;

function TEbookFile.GetTitleSort: string;
begin
  if not fMetadataLoaded then
    LoadMetadata;
  Result := fTitleSort;
end;

procedure TEbookFile.Assign(Source: TEbookFile);
begin
  ClearMetadata;
  fTitle := Source.Title;
  fTitleSort := Source.TitleSort;
  fAuthor := Source.Author;
  fAuthorSort := Source.AuthorSort;
  fSeries := Source.Series;
  fSeriesIndex := Source.SeriesIndex;
  fLanguage := Source.Language;
  fSize := Source.Size;
  fCreationDate := Source.CreationDate;
  fModificationDate := Source.ModificationDate;
  fAccessDate := Source.AccessDate;
  if Source.Cover <> nil then
    begin
    fCover := TGraphicClass(Source.Cover.ClassType).Create;
    fCover.Assign(Source.Cover);
    end;
end;

procedure TEbookFile.ClearMetadata;
begin
  fTitle := '';
  fTitleSort := '';
  fAuthor := '';
  fAuthorSort := '';
  fSeries := '';
  fSeriesIndex := -1;
  fLanguage := '';
  fSize := -1;
  fCreationDate := 0;
  fModificationDate := 0;
  fAccessDate := 0;
  FreeAndNil(fCover);
end;

procedure TEbookFile.LoadMetaData;
begin
  ClearMetadata;
  fSize := GetFileSize(FileName);
  GetFileDateTime(FileName, fCreationDate, fModificationDate, fAccessDate);
  fMetadataLoaded := True;
end;

procedure TEbookFile.Refresh;
begin
  LoadMetaData;
end;

procedure TEbookFile.SetAuthor(const Value: string);
begin
  fAuthor := Value;
  fMetadataLoaded := True;
end;

procedure TEbookFile.SetAuthorSort(const Value: string);
begin
  fAuthorSort := Value;
  fMetadataLoaded := True;
end;

procedure TEbookFile.SetCover(const Value: TGraphic);
begin
  fCover := Value;
  fMetadataLoaded := True;
end;

procedure TEbookFile.SetLanguage(const Value: string);
begin
  fLanguage := Value;
  fMetadataLoaded := True;
end;

procedure TEbookFile.SetSeries(const Value: string);
begin
  fSeries := Value;
  fMetadataLoaded := True;
end;

procedure TEbookFile.SetSeriesIndex(const Value: integer);
begin
  fSeriesIndex := Value;
  fMetadataLoaded := True;
end;

procedure TEbookFile.SetTitle(const Value: string);
begin
  fTitle := Value;
  fMetadataLoaded := True;
end;

procedure TEbookFile.SetTitleSort(const Value: string);
begin
  fTitleSort := Value;
  fMetadataLoaded := True;
end;

initialization
  RegisteredEbookClasses := TStringList.Create;
  RegisteredEbookClasses.Sorted := True;

finalization
  FreeAndNil(RegisteredEbookClasses);

end.
