(******************************************************************************

______________________________________________________________________________

ColBuild v0.10                                               (c) 2009-12 Pepak
http://www.pepak.net/e-books/prs-505-sbirky/              http://www.pepak.net
______________________________________________________________________________


Copyright (c) 2009-12 Pepak (http://www.pepak.net)
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of Pepak nor the
      names of his contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL PEPAK BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

******************************************************************************)

unit uEpubEbooks;

interface

uses
  SysUtils, Classes, Windows, Graphics, Jpeg, PngImage, GifImg, SciZipFile,
  uEbooks, uXml;

type
  TEpubFile = class(TEbookFile)
  private
  protected
    function FindFileInZip(const Zip: TZipFile; const FileName: string; out Index: integer; CaseSensitive: boolean = False): boolean;
    function FindOPFFile(const XmlData: AnsiString; out FileName: string): boolean;
    function GetMetadataFromOPF(const XmlData: AnsiString; out CoverFile, CoverType: string): boolean;
    function GetCover(const CoverData: AnsiString; const CoverType: string): boolean;
    procedure LoadMetaData; override;
  public
    class function IsSupportedFile(const AFileName: string): boolean; override;
    constructor Create(const AFileName: string); override;
  end;

implementation

{ TEpubFile }

class function TEpubFile.IsSupportedFile(const AFileName: string): boolean;
const
  Supported: array[0..0] of string = ('epub');
begin
  Result := IsSupportedFileByExtension(AFileName, Supported);
end;

constructor TEpubFile.Create(const AFileName: string);
begin
  inherited;
end;

function TEpubFile.FindFileInZip(const Zip: TZipFile; const FileName: string; out Index: integer; CaseSensitive: boolean): boolean;

  function NormalizeFileName(const FileName: string; CaseSensitive: boolean): string;
    var
      i: Integer;
    begin
      Result := FileName;
      if Result = '' then
        Exit;
      for i := 1 to Length(Result) do
        if Result[i] = '\' then
          Result[i] := '/';
      while Result[1] = '/' do
        Delete(Result, 1, 1);
      i := Length(Result);
      while (i > 0) and (Result[i] = '/') do
        Dec(i);
      SetLength(Result, i);
      if not CaseSensitive then
        Result := AnsiUpperCase(Result);
    end;

var
  FNWant: string;
  i: integer;
begin
  Result := False;
  Index := -1;
  FNWant := NormalizeFileName(FileName, CaseSensitive);
  for i := 0 to Pred(Zip.Count) do
    if NormalizeFileName(string(Zip.Name[i]), CaseSensitive) = FNWant then
      begin
      Index := i;
      Result := True;
      Break;
      end;
end;

function TEpubFile.FindOPFFile(const XmlData: AnsiString; out FileName: string): boolean;
var
  Xml: TXmlDoc;
  Node: TXmlNode;
begin
  Result := False;
  FileName := '';
  Xml := TXmlDoc.Create;
  try
    Xml.LoadFromBinaryString(XmlData);
    if XmlNodeByPathAndAttr(Xml, 'rootfiles/rootfile', 'media-type', 'application/oebps-package+xml', Node) then
      begin
      FileName := Node.AttributeByNameWide['full-path'];
      Result := FileName <> '';
      end;
  finally
    FreeAndNil(Xml);
    end;
end;

function TEpubFile.GetMetadataFromOPF(const XmlData: AnsiString; out CoverFile, CoverType: string): boolean;
var
  Xml: TXmlDoc;
  Meta, Node: TXmlNode;
  s, NS_DC, NS_OPF: string;
begin
  Result := False;
  CoverFile := '';
  CoverType := '';
  Xml := TXmlDoc.Create;
  try
    Xml.LoadFromBinaryString(XmlData);
    if XmlNodeByPath(Xml, 'metadata', Meta) then
      begin
      Result := True;
      NS_DC := 'dc:'; {$MESSAGE WARN 'Tohle je blbe, melo by se pouzit odkazovani pres namespace URI'}
      NS_OPF := 'opf:'; {$MESSAGE WARN 'Tohle je blbe, melo by se pouzit odkazovani pres namespace URI'}
      Title := XmlValueByPath(Meta, NS_DC + 'title');
      if XmlNodeByPathAndAttr(Meta, 'meta', 'name', 'calibre:title_sort', Node) then
        TitleSort := Node.AttributeByNameWide['content'];
      if XmlNodeByPathAndAttr(Meta, NS_DC + 'creator', NS_OPF + 'role', 'aut', Node) then
        begin
        Author := XmlValueByPath(Node, '');
        AuthorSort := Node.AttributeByNameWide[UTF8String(NS_OPF) + 'file-as'];
        end;
      if XmlNodeByPathAndAttr(Meta, 'meta', 'name', 'calibre:series', Node) then
        Series := Node.AttributeByNameWide['content'];
      if XmlNodeByPathAndAttr(Meta, 'meta', 'name', 'calibre:series_index', Node) then
        SeriesIndex := StrToIntDef(Node.AttributeByNameWide['content'], -1);
      Language := XmlValueByPath(Meta, NS_DC + 'language');
      if XmlNodeByPathAndAttr(Meta, 'meta', 'name', 'cover', Node) then
        begin
        s := Node.AttributeByNameWide['content'];
        if s <> '' then
          if XmlNodeByID(Xml, s, Node) then
            begin
            CoverFile := Node.AttributeByNameWide['href'];
            CoverType := Node.AttributeByNameWide['media-type'];
            end;
        end;
      end;
  finally
    FreeAndNil(Xml);
    end;
end;

function TEpubFile.GetCover(const CoverData: AnsiString; const CoverType: string): boolean;
type
  TMimeTypeToGraphic = record
    MimeType: string;
    GraphicClass: TGraphicClass;
    end;
const
  MimeTypeToGraphic: array [0..2] of TMimeTypeToGraphic = (
    (MimeType: 'image/gif'; GraphicClass: TGifImage),
    (MimeType: 'image/jpeg'; GraphicClass: TJpegImage),
    (MimeType: 'image/png'; GraphicClass: TPngImage)
  );
var
  i: integer;
  S: TMemoryStream;
begin
  Result := False;
  Cover.Free;
  SetCover(nil);
  for i := 0 to Pred(Length(MimeTypeToGraphic)) do
    if AnsiCompareText(CoverType, MimeTypeToGraphic[i].MimeType) = 0 then
      begin
      S := TMemoryStream.Create;
      try
        S.WriteBuffer(CoverData[1], Length(CoverData));
        S.Position := 0;
        SetCover(MimeTypeToGraphic[i].GraphicClass.Create);
        Cover.LoadFromStream(S);
        Result := True;
        Break;
      finally
        FreeAndNil(S);
        end;
      end;
end;

procedure TEpubFile.LoadMetaData;
var
  Zip: TZipFile;
  ix: integer;
  FN, CoverType: string;
begin
  inherited LoadMetaData;
  Zip := TZipFile.Create;
  try
    Zip.LoadFromFile(FileName);
    if FindFileInZip(Zip, 'META-INF/container.xml', ix, True) then
      if FindOPFFile(Zip.Data[ix], FN) then
        if FindFileInZip(Zip, FN, ix, True) then
          if GetMetadataFromOPF(Zip.Data[ix], FN, CoverType) then
            if FN <> '' then
              if FindFileInZip(Zip, FN, ix, True) then
                GetCover(Zip.Data[ix], CoverType);
  finally
    FreeAndNil(Zip);
    end;
end;

initialization
  TEbookFileList.RegisterEbookClass(TEpubFile, '');

end.
