(******************************************************************************

______________________________________________________________________________

ColBuild v0.10                                               (c) 2009-12 Pepak
http://www.pepak.net/e-books/prs-505-sbirky/              http://www.pepak.net
______________________________________________________________________________


Copyright (c) 2009-12 Pepak (http://www.pepak.net)
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of Pepak nor the
      names of his contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL PEPAK BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

******************************************************************************)

unit uMain;
{$INCLUDE 'ColBuild.inc'}

interface

uses
  SysUtils, Classes, Windows,
  uConsoleApp, uCompatibility, uSystem, uXml,
  uEbooks, uDevices, uMedia;

type
  TMetadataHandling = (mhNone, mhCreate, mhCreateAndUpdate);

  TColBuildApp = class(TConsoleApp)
    private
      fConfig: TXmlDoc;
      fCollectionDefFileNames: TStringList;
      fReaderRoot: string;
      fDeviceType: string;
      fDevice: TDevice;
      fMediaType: TMediaType;
      fEbookList: TEbookFileList;
      fQuiet: boolean;
      fMetadataHandling: TMetadataHandling;
      fCreateCollections: boolean;
      fGenerateDummyCovers: boolean;
{
      fSkipMetadata: boolean;
      fSkipMetadataUpdate: boolean;
      fSkipCollections: boolean;
}
    protected // Application info
      function AppTitle: string; override;
      function AppVersion: string; override;
      function DoExecute: integer; override;
      procedure ShowSyntax(const Error: string = ''); override;
      procedure ShowSupportedDevices;
      procedure ShowSupporttedMedia;
      procedure ParamInitialize; override;
      function ParseCommandLine: integer;
      procedure LoadConfig(Silent: boolean = False);
      procedure LoadEbooks(const RootDir: string);
      function ProcessWildCardFile(const FileName: string; const SearchRec: TSearchRec): boolean; override;
      function RegisterBooks: boolean;
      function BuildCollection(const FileName: string; const ReaderPathPrefix: string = ''): boolean;
      function FileNameRelativeTo(const FileName, RelativeToFileName: string): string;
      property Config: TXmlDoc read fConfig;
      property EbookList: TEbookFileList read fEbookList;
      property Device: TDevice read fDevice;
      property CollectionDefFileNames: TStringList read fCollectionDefFileNames;
    public
      constructor Create; override;
      destructor Destroy; override;
      property ReaderRoot: string read fReaderRoot write fReaderRoot;
      property DeviceType: string read fDeviceType write fDeviceType;
      property MediaType: TMediaType read fMediaType write fMediaType;
      property Quiet: boolean read fQuiet write fQuiet;
      property MetadataHandling: TMetadataHandling read fMetadataHandling write fMetadataHandling;
      property CreateCollections: boolean read fCreateCollections write fCreateCollections;
      property GenerateDummyCovers: boolean read fGenerateDummyCovers write fGenerateDummyCovers;
{
      property SkipMetadata: boolean read fSkipMetadata write fSkipMetadata;
      property SkipMetadataUpdate: boolean read fSkipMetadataUpdate write fSkipMetadataUpdate;
      property SkipCollections: boolean read fSkipCollections write fSkipCollections;
}
    end;

const
  RESCODE_HELP = 1;
  RESCODE_BAD_PARAMS = 2;
  RESCODE_REGISTERBOOKS = 3;
  RESCODE_DELETECOLLECTIONS = 4;
  RESCODE_CREATECOLLECTIONS = 5;

procedure Main;

implementation

procedure Main;
begin
  ExitCode := ExecuteConsoleApp(TColBuildApp);
end;

const
  colorFile1 = ccLightCyan;
  colorFile2 = ccWhite;

{ TColBuildApp }

function TColBuildApp.AppTitle: string;
begin
  Result := 'ColBuild';
end;

function TColBuildApp.AppVersion: string;
begin
  Result := {$INCLUDE 'ColBuild.version'} ;
end;

constructor TColBuildApp.Create;
begin
  inherited;
  fConfig := TXmlDoc.Create;
  fEbookList := TEbookFileList.Create(True);
  fCollectionDefFileNames := TStringList.Create;
  LoadConfig(True);
  ParamInitialize;
end;

destructor TColBuildApp.Destroy;
begin
  FreeAndNil(fEbookList);
  FreeAndNil(fDevice);
  FreeAndNil(fCollectionDefFileNames);
  FreeAndNil(fConfig);
  inherited;
end;

procedure TColBuildApp.LoadConfig;
var
  FN: string;
begin
  try
    FN := ChangeFileExt(ParamStr(0), '.xml');
    if FileExists(FN) then
      Config.LoadFromFile(FN);
  except
    FreeAndNil(fConfig);
    fConfig := TXmlDoc.Create;
    if not Silent then
      Raise;
    end;
end;

procedure TColBuildApp.ShowSupportedDevices;
var
  L: TStringList;
  i: integer;
begin
  L := TStringList.Create;
  try
    TDevice.ListDevices(L);
    for i := 0 to Pred(L.Count) do
      begin
      Write('               ');
      WriteColored(ccWhite, L[i]);
      Writeln(': ' + TDeviceClass(L.Objects[i]).Description);
      end;
  finally
    FreeAndNil(L);
    end;
end;

procedure TColBuildApp.ShowSupporttedMedia;
var
  mt: TMediaType;
begin
  for mt := Low(TMediaType) to High(TMediaType) do
    begin
    Write('               ');
    WriteColored(ccWhite, MediaTypeInfo[mt].ShortID);
    Write(', ');
    WriteColored(ccWhite, MediaTypeInfo[mt].LongID);
    Writeln(': ' + MediaTypeInfo[mt].Description);
    end;
end;

procedure TColBuildApp.ParamInitialize;
var
  s: string;
begin
  inherited;
  if Config <> nil then
    begin
    ReaderRoot := XmlValueByPath(Config, 'config/reader_root', '.');
    DeviceType := XmlValueByPath(Config, 'config/device_type', '');
    MediaType := StrToMediaTypeDef(XmlValueByPath(Config, 'config/media_type', ''), Low(TMediaType));
    Quiet := StrToIntDef(XmlValueByPath(Config, 'config/quiet', ''), 0) <> 0;
    s := XmlValueByPath(Config, 'config/metadata_handling', '');
    if AnsiCompareText(s, 'none') = 0 then
      MetadataHandling := mhNone
    else if AnsiCompareText(s, 'create') = 0 then
      MetadataHandling := mhCreate
    else if AnsiCompareText(s, 'update') = 0 then
      MetadataHandling := mhCreateAndUpdate
    else
      MetadataHandling := mhCreateAndUpdate;
    CreateCollections := StrToIntDef(XmlValueByPath(Config, 'config/create_collections', ''), 1) <> 0;
    GenerateDummyCovers := StrToIntDef(XmlValueByPath(Config, 'config/generate_dummy_covers', ''), 0) <> 0;
    CollectionDefFileNames.Clear;
    end;
end;

procedure TColBuildApp.ShowSyntax(const Error: string);
begin
  inherited;
  WriteColored(ccWhite, '<arg> [<arg>] ...'); Writeln;
  Writeln;
  WriteColored(ccWhite, ' -h, -?'); Writeln(' ...... Show this help screen.');
  WriteColored(ccWhite, ' -d <name>'); Writeln(' ... Use driver for device <name>. Supported devices: '); ShowSupportedDevices;
  WriteColored(ccWhite, ' -m <type>'); Writeln(' ... Assume destination media <type>. Supported: '); ShowSupporttedMedia;
  WriteColored(ccWhite, ' -r <dir>'); Writeln(' .... Set device''s root to <dir>. Default: current working directory.');
  WriteColored(ccWhite, ' -q[-]'); Writeln(' ....... [Don''t] Use quiet mode (hide progress). Default: don''t use.');
  WriteColored(ccWhite, ' -mu'); Writeln(' ......... Create and update metadata for all books. This is default.');
  WriteColored(ccWhite, ' -mc'); Writeln(' ......... Create matadata for new books but don''t update old ones.');
  WriteColored(ccWhite, ' -mn'); Writeln(' ......... Don''t work with metadata, except when needed for collections.');
  WriteColored(ccWhite, ' -dc[-]'); Writeln(' ...... [Don''t] Generate dummy covers where real covers are missing.');
  WriteColored(ccWhite, ' -c[-]'); Writeln(' ....... [Don''t] Create collections. Default: create.');
  WriteColored(ccWhite, ' <file>'); Writeln(' ...... Build collections defined in file <file>. Needed if -c is used.');
end;

function TColBuildApp.ParseCommandLine: integer;
var
  Param: string;
  mt: TMediaType;
begin
  Result := RESCODE_OK;
  FreeAndNil(fDevice);
  if ParamCount = 0 then
    begin
    ShowSyntax;
    Result := RESCODE_HELP;
    Exit;
    end
  else
    begin
    ParamInitialize;
    while ParamGetNext(Param) do
      if Param[1] = '-' then
        begin
        if (AnsiCompareText(Param, '-?') = 0) or (AnsiCompareText(Param, '-h') = 0) then
          begin
          ShowSyntax;
          Result := RESCODE_HELP;
          Exit;
          end
        else if AnsiCompareText(Param, '-r') = 0 then
          if ParamGetNext(Param) then
            ReaderRoot := Param
          else
            begin
            ShowSyntax('Need a directory for parameter %s', [Param]);
            Result := RESCODE_BAD_PARAMS;
            Exit;
            end
        else if AnsiCompareText(Param, '-d') = 0 then
          if ParamGetNext(Param) then
            DeviceType := Param
          else
            begin
            ShowSyntax('Need a device name for parameter %s', [Param]);
            Result := RESCODE_BAD_PARAMS;
            Exit;
            end
        else if AnsiCompareText(Param, '-m') = 0 then
          if ParamGetNext(Param) then
            if StrToMediaType(Param, mt) then
              MediaType := mt
            else
              begin
              ShowSyntax('Unknown media type %s', [Param]);
              Result := RESCODE_BAD_PARAMS;
              Exit;
              end
          else
            begin
            ShowSyntax('Need a media type for parameter %s', [Param]);
            Result := RESCODE_BAD_PARAMS;
            Exit;
            end
        else if AnsiCompareText(Param, '-q') = 0 then
          Quiet := True
        else if AnsiCompareText(Param, '-q-') = 0 then
          Quiet := False
        else if AnsiCompareText(Param, '-mn') = 0 then
          MetadataHandling := mhNone
        else if AnsiCompareText(Param, '-mc') = 0 then
          MetadataHandling := mhCreate
        else if AnsiCompareText(Param, '-mu') = 0 then
          MetadataHandling := mhCreateAndUpdate
        else if AnsiCompareText(Param, '-c') = 0 then
          CreateCollections := True
        else if AnsiCompareText(Param, '-c-') = 0 then
          CreateCollections := False
        else if AnsiCompareText(Param, '-dc') = 0 then
          GenerateDummyCovers := True
        else if AnsiCompareText(Param, '-dc-') = 0 then
          GenerateDummyCovers := False
        else
          begin
          ShowSyntax('Unknown parameter %s', [Param]);
          Result := RESCODE_BAD_PARAMS;
          Exit;
          end;
        end
      else
        CollectionDefFileNames.Add(Param);
    end;
  ReaderRoot := AddLastSlash(ReaderRoot);
  if DeviceType = '' then
    begin
    ShowSyntax('Need device driver.');
    Result := RESCODE_BAD_PARAMS;
    Exit;
    end;
  if not TDevice.CreateDevice(DeviceType, ReaderRoot, fDevice) then
    begin
    ShowSyntax('Unknown device driver "%s".', [DeviceType]);
    Result := RESCODE_BAD_PARAMS;
    Exit;
    end;
  {
  if SkipMetadata and SkipCollections then
    begin
    ShowSyntax('Nothing to do.');
    Result := RESCODE_BAD_PARAMS;
    Exit;
    end;
  }
end;

function TColBuildApp.DoExecute: integer;
var
  i: integer;
begin
  Result := ParseCommandLine;
  if Result = RESCODE_OK then
    begin
    Device.MediaType := MediaType;
    EbookList.Clear;
    EbookList.SkipMetadataExtraction := MetadataHandling = mhNone;
    EbookList.GenerateDummyCovers := GenerateDummyCovers;
    Writeln('Loading e-books...');
    LoadEbooks(ReaderRoot);
    EbookList.Sort;
    Writeln('Accessing reader device...');
    Device.Connect;
    if MetadataHandling <> mhNone then
      begin
      Writeln('Registering e-books with the device...');
      if not RegisterBooks then
        begin
        Result := RESCODE_REGISTERBOOKS;
        Exit;
        end;
      end;
    if CreateCollections then
      begin
      Writeln('Deleting old collections...');
      if not Device.DeleteAllCollections then
        begin
        Result := RESCODE_DELETECOLLECTIONS;
        Exit;
        end;
      Writeln('Building new collections...');
      for i := 0 to Pred(CollectionDefFileNames.Count) do
        if not BuildCollection(CollectionDefFileNames[i]) then
          begin
          Result := RESCODE_CREATECOLLECTIONS;
          Exit;
          end;
      end;
    Device.Commit;
    Writeln('Done.');
    end;
end;

function TColBuildApp.FileNameRelativeTo(const FileName, RelativeToFileName: string): string;
begin
  Result := FileName;
  if Result <> '' then
    if Pos(DriveDelim, Result) <= 0 then
      if Copy(Result, 1, 2) <> '\\' then
        if Result[1] = '\' then
          Result := ExtractFileDrive(RelativeToFileName) + Result
        else
          Result := ExtractFilePath(RelativeToFileName) + Result;
end;

procedure TColBuildApp.LoadEbooks(const RootDir: string);
begin
  ProcessWildCard(RootDir + '*.*', True);
end;

function TColBuildApp.ProcessWildCardFile(const FileName: string; const SearchRec: TSearchRec): boolean;
begin
  if not Device.IsSupportedFile(FileName) then
    Result := True
  else
    try
      if not Quiet then
        begin
        WriteColored(colorFile2, FileName);
        Writeln;
        end;
      Result := EbookList.AddFile(FileName) >= 0;
      if not Quiet then
        if not Result then
          begin
          WriteColored(ccYellow, 'unknown file type');
          Writeln;
          end;
    except
      on E: Exception do
        begin
        ShowError('%s: %s', [E.ClassName, E.Message]);
        Result := False;
        end;
      end;
end;

function TColBuildApp.RegisterBooks: boolean;
var
  i: integer;
  OK: boolean;
begin
  Result := True;
  for i := 0 to Pred(EbookList.Count) do
    try
      if (MetadataHandling <> mhCreateAndUpdate) and Device.IsRegisteredBook(EbookList[i]) then
        begin
        {
        if not Quiet then
          begin
          WriteColored(ccYellow, 'Skipped');
          Writeln;
          end;
        }
        end
      else
        begin
        if not Quiet then
          begin
          Write('Registering ');
          WriteColored(colorFile2, EbookList[i].FileName);
          Writeln;
          end;
        OK := Device.RegisterBook(EbookList[i]);
        if not OK then
          Result := False;
        if not Quiet then
          if not OK then
            begin
            WriteColored(ccYellow, 'Registration failed');
            Writeln;
            end;
        end;
    except
      on E: Exception do
        begin
        ShowError('%s: %s', [E.ClassName, E.Message]);
        Result := False;
        end;
      end;
end;

function TColBuildApp.BuildCollection(const FileName: string; const ReaderPathPrefix: string): boolean;

  function SlashToPathDelimiter(const s: string): string;
    begin
      Result := StringReplace(s, '/', PathDelim, [rfReplaceAll]);
    end;

  function GetFilePrefix(Node: TXmlNode; const CurrentPrefix: string): string;
    begin
      if XmlNodeByPath(Node, 'file_prefix', Node) then
        if Node.HasAttribute('replace') and (StrToIntDef(Node.AttributeByNameWide['replace'], 0) <> 0) then
          Result := XmlValueIncludingCData(Node)
        else
          Result := CurrentPrefix + XmlValueIncludingCData(Node)
      else
        Result := CurrentPrefix;
    end;

var
  Xml: TXmlDoc;
  Node, Files, AllFiles: TXmlNode;
  i, j, k: integer;
  FN, FN2, Prefix, Name, NameSort, ColPrefix: string;
  Collection: TEbookFileList;
begin
  try
    if not Quiet then
      begin
      Write('Building collections from ');
      WriteColored(colorFile1, FileName);
      Writeln;
      end;
    Collection := TEbookFileList.Create(False);
    try
      Xml := TXmlDoc.Create;
      try
        Xml.LoadFromFile(FileName);
        Result := True;
        Prefix := GetFilePrefix(Xml.Root, ReaderPathPrefix);
        for i := 0 to Pred(Xml.Root.NodeCount) do
          try
            Node := Xml.Root.Nodes[i];
            if Node.Name = 'include' then
              begin
              FN := SlashToPathDelimiter(XmlValueIncludingCData(Node));
              FN := FileNameRelativeTo(FN, FileName);
              if not BuildCollection(FN, Prefix) then
                Result := False;
              end
            else if Node.Name = 'collection' then
              try
                Name := XmlValueByPath(Node, 'name');
                NameSort := XmlValueByPath(Node, 'name_sort', Name);
                if not Quiet then
                  begin
                  Write('Collection ');
                  WriteColored(colorFile1, Name);
                  Writeln;
                  end;
                ColPrefix := GetFilePrefix(Node, Prefix);
                Collection.Clear;
                if XmlNodeByPath(Node, 'files', Files) then
                  for j := 0 to Pred(Files.NodeCount) do
                    if Files.Nodes[j].Name = 'file' then
                      begin
                      FN := SlashToPathDelimiter(ColPrefix + XmlValueIncludingCData(Files.Nodes[j]));
                      for k := 0 to Pred(EbookList.Count) do
                        begin
                        FN2 := SlashToPathDelimiter(Device.FindFilePathRelativeToRoot(EbookList[k]));
                        if (AnsiCompareText(FN, FN2) = 0) or (AnsiCompareText(FN, ChangeFileExt(FN2, '')) = 0) then
                          begin
                          Collection.Add(EbookList[k]);
                          Break;
                          end;
                        end;
                      end;
                if XmlNodeByPath(Node, 'all_files', AllFiles) then
                  begin
                  FN := SlashToPathDelimiter(ColPrefix + XmlValueIncludingCData(AllFiles));
                  for k := 0 to Pred(EbookList.Count) do
                    begin
                    FN2 := SlashToPathDelimiter(Device.FindFilePathRelativeToRoot(EbookList[k]));
                    if (AnsiCompareText(FN, Copy(FN2, 1, Length(FN))) = 0) then
                      Collection.Add(EbookList[k]);
                    end;
                  end;
                if Collection.Count > 0 then
                  begin
                  if not Quiet then
                    for j := 0 to Pred(Collection.Count) do
                      begin
                      WriteColored(colorFile2, Collection[j].FileName);
                      Writeln;
                      end;
                  if not Device.CreateCollection(Name, NameSort, Collection) then
                    begin
                    Result := False;
                    if not Quiet then
                      begin
                      WriteColored(ccYellow, 'FAILED');
                      Writeln;
                      end;
                    end;
                  end;
              except
                on E: Exception do
                  begin
                  ShowError('%s: %s', [E.ClassName, E.Message]);
                  Result := False;
                  end;
              end;
          except
            on E: Exception do
              begin
              ShowError('%s: %s', [E.ClassName, E.Message]);
              Result := False;
              end;
          end;
      finally
        FreeAndNil(Xml);
        end;
    finally
      FreeAndNil(Collection);
      end;
  except
    on E: Exception do
      begin
      ShowError('%s: %s', [E.ClassName, E.Message]);
      Result := False;
      end;
    end;
end;

end.
