(******************************************************************************

______________________________________________________________________________

ColBuild v0.10                                               (c) 2009-12 Pepak
http://www.pepak.net/e-books/prs-505-sbirky/              http://www.pepak.net
______________________________________________________________________________


Copyright (c) 2009-12 Pepak (http://www.pepak.net)
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of Pepak nor the
      names of his contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL PEPAK BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

******************************************************************************)

unit uPRS_505;

interface

uses
  SysUtils, Classes, Windows, Graphics, Jpeg, SynaCode,
  uDevices, uMedia, uEbooks, uSystem, uGuid, uXml;

type
  TSonyPRS505 = class(TDevice)
    private
      fDatabase: TXmlDoc;
      fDatabaseFileName: string;
    protected
      procedure CreateNewDatabase;
      function NextId: integer;
      function GetNextId(Node: TXmlNode): integer;
      function BookTag: string;
      function PlaylistTag: string;
      function PlaylistItemTag: string;
      function ThumbnailTag: string;
      function ImageTag: string;
      function BookDateTime(Ebook: TEbookFile): TDateTime;
      function DateTime2Xml(Value: TDateTime): string;
      function FindBookNode(Ebook: TEbookFile; out Node: TXmlNode): boolean;
      function GetMimeType(Ebook: TEbookFile): string;
      function UpdateBook(Ebook: TEbookFile; Node: TXmlNode): boolean;
      function CreateBook(Ebook: TEbookFile; out Node: TXmlNode): boolean;
      function UpdateBookThumbnail(Ebook: TEbookFile; Node: TXmlNode): boolean;
      function SupportedMediaTypes: TMediaTypes; override;
      function SourceID: string;
      property Database: TXmlDoc read fDatabase;
      property DatabaseFileName: string read fDatabaseFileName;
    public
      class function Description: string; override;
      constructor Create(const ARootDir: string); override;
      destructor Destroy; override;
      procedure Connect; override;
      procedure Disconnect; override;
      procedure Commit; override;
      procedure Rollback; override;
      function IsSupportedFile(const AFileName: string): boolean; override;
      function IsRegisteredBook(Ebook: TEbookFile): boolean; override;
      function DeleteAllCollections: boolean; override;
      function RegisterBook(Ebook: TEbookFile): boolean; override;
      function CreateCollection(const Name, NameSort: string; Ebooks: TEbookFileList): boolean; override;
    end;

implementation

{$R 'Resources\PRS_505_Internal.res'}
{$R 'Resources\PRS_505_SD.res'}
{$R 'Resources\PRS_505_MS.res'}

const
  FILENAME_BOOKSDB_INTERNAL = 'database\cache\media.xml';
  FILENAME_BOOKSDB_SDCARD = 'Sony Reader\database\cache.xml';
  FILENAME_BOOKSDB_MSCARD = 'Sony Reader\database\cache.xml';

const
  BOOKTAG_INTERNAL = 'xs1:text';
  BOOKTAG_SDCARD = 'text';
  BOOKTAG_MSCARD = 'text';
  PLAYLISTTAG_INTERNAL = 'xs1:playlist';
  PLAYLISTTAG_SDCARD = 'playlist';
  PLAYLISTTAG_MSCARD = 'playlist';
  PLAYLISTITEMTAG_INTERNAL = 'xs1:item';
  PLAYLISTITEMTAG_SDCARD = 'item';
  PLAYLISTITEMTAG_MSCARD = 'item';
  THUMBNAILTAG_INTERNAL = 'xs1:thumbnail';
  THUMBNAILTAG_SDCARD = 'thumbnail';
  THUMBNAILTAG_MSCARD = 'thumbnail';
  IMAGETAG_INTERNAL = 'xs1:jpeg';
  IMAGETAG_SDCARD = 'jpeg';
  IMAGETAG_MSCARD = 'jpeg';
  RECORDSTAG = 'records';

const
  THUMB_WIDTH = 51;
  THUMB_HEIGHT = 68;

type
  TMimeType = record
    Extension, MimeType: string;
    end;
const
  MimeTypes: array[0..4] of TMimeType = (
    (Extension: '.epub'; MimeType: 'application/epub+zip'),
    (Extension: '.txt'; MimeType: 'text/plain'),
    (Extension: '.pdf'; MimeType: 'application/pdf'),
    (Extension: '.lrf'; MimeType: 'application/x-sony-bbeb'),
    (Extension: '.rtf'; MimeType: 'application/rtf')
  );

{ TSonyPRS505 }

class function TSonyPRS505.Description: string;
begin
  Result := 'Sony Reader PRS-505';
end;

constructor TSonyPRS505.Create(const ARootDir: string);
begin
  inherited;
end;

destructor TSonyPRS505.Destroy;
begin
  inherited;
end;

procedure TSonyPRS505.Connect;
begin
  if not Connected then
    begin
    CreateNewDatabase;
    end;
  inherited;
end;

procedure TSonyPRS505.Disconnect;
begin
  if Connected then
    begin
    Rollback;
    FreeAndNil(fDatabase);
    end;
  inherited;
end;

procedure TSonyPRS505.Commit;
begin
  ForceDirectories(ExtractFilePath(DatabaseFileName));
  Database.SaveToFile(DatabaseFileName);
  inherited;
end;

procedure TSonyPRS505.Rollback;
begin
  if FileExists(DatabaseFileName) then
    Database.LoadFromFile(DatabaseFileName)
  else
    CreateNewDatabase;
  inherited;
end;

procedure TSonyPRS505.CreateNewDatabase;
var
  DBStream: TStream;
begin
  FreeAndNil(fDatabase);
  fDatabaseFileName := '';
  DBStream := nil;
  try
    fDatabase := TXmlDoc.Create;
    try
      case MediaType of
        mtInternalMemory:
          begin
            fDatabaseFileName := FindFilePathInReader(FILENAME_BOOKSDB_INTERNAL);
            DBStream := TResourceStream.Create(hInstance, 'PRS_505_Internal', 'BINARY');
          end;
        mtSDcard:
          begin
            fDatabaseFileName := FindFilePathInReader(FILENAME_BOOKSDB_SDCARD);
            DBStream := TResourceStream.Create(hInstance, 'PRS_505_SD', 'BINARY');
          end;
        mtMScard:
          begin
            fDatabaseFileName := FindFilePathInReader(FILENAME_BOOKSDB_MSCARD);
            DBStream := TResourceStream.Create(hInstance, 'PRS_505_MS', 'BINARY');
          end;
      end;
      if FileExists(DatabaseFileName) then
        Database.LoadFromFile(DatabaseFileName)
      else
        fDatabase.LoadFromStream(DBStream);
    except
      FreeAndNil(fDatabase);
      Raise;
      end;
  finally
    FreeAndNil(DBStream);
    end;
end;

function TSonyPRS505.IsSupportedFile(const AFileName: string): boolean;
var
  Ext: string;
  i: integer;
begin
  Result := False;
  Ext := ExtractFileExt(AFileName);
  for i := 0 to Pred(Length(MimeTypes)) do
    if AnsiCompareText(Ext, MimeTypes[i].Extension) = 0 then
      begin
      Result := True;
      Break;
      end;
end;

function TSonyPRS505.BookTag: string;
begin
  case MediaType of
    mtInternalMemory:
      Result := BOOKTAG_INTERNAL;
    mtSDcard:
      Result := BOOKTAG_SDCARD;
    mtMScard:
      Result := BOOKTAG_MSCARD;
    else
      Result := '';
    end;
end;

function TSonyPRS505.PlaylistItemTag: string;
begin
  case MediaType of
    mtInternalMemory:
      Result := PLAYLISTITEMTAG_INTERNAL;
    mtSDcard:
      Result := PLAYLISTITEMTAG_SDCARD;
    mtMScard:
      Result := PLAYLISTITEMTAG_MSCARD;
    else
      Result := '';
    end;
end;

function TSonyPRS505.PlaylistTag: string;
begin
  case MediaType of
    mtInternalMemory:
      Result := PLAYLISTTAG_INTERNAL;
    mtSDcard:
      Result := PLAYLISTTAG_SDCARD;
    mtMScard:
      Result := PLAYLISTTAG_MSCARD;
    else
      Result := '';
    end;
end;

function TSonyPRS505.ThumbnailTag: string;
begin
  case MediaType of
    mtInternalMemory:
      Result := THUMBNAILTAG_INTERNAL;
    mtSDcard:
      Result := THUMBNAILTAG_SDCARD;
    mtMScard:
      Result := THUMBNAILTAG_MSCARD;
    else
      Result := '';
    end;
end;

function TSonyPRS505.ImageTag: string;
begin
  case MediaType of
    mtInternalMemory:
      Result := IMAGETAG_INTERNAL;
    mtSDcard:
      Result := IMAGETAG_SDCARD;
    mtMScard:
      Result := IMAGETAG_MSCARD;
    else
      Result := '';
    end;
end;

function TSonyPRS505.SourceID: string;
begin
  Result := '1';
end;

function TSonyPRS505.BookDateTime(Ebook: TEbookFile): TDateTime;
begin
  case MediaType of
    mtInternalMemory:
      Result := Ebook.CreationDate;
    mtSDcard:
      Result := Ebook.ModificationDate;
    mtMScard:
      Result := Ebook.ModificationDate;
    else
      Result := 0;
    end;
end;

function TSonyPRS505.DateTime2Xml(Value: TDateTime): string;
const
  Days: array[1..7] of string = ('Sun', 'Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat');
  Months: array[1..12] of string = ('Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec');
var
  D, M, Y, H, Mn, S, MS: Word;
begin
  DecodeDate(Value, Y, M, D);
  DecodeTime(Value, H, Mn, S, MS);
  Result := Format('%s, %02.2d %s %04.4d %02.2d:%02.2d:%02.2d GMT', [Days[DayOfWeek(Value)], D, Months[M], Y, H, Mn, S]);
end;

function TSonyPRS505.FindBookNode(Ebook: TEbookFile; out Node: TXmlNode): boolean;
var
  FN: string;
begin
  Connect;
  FN := FindFilePathRelativeToRoot(Ebook);
  Result := XmlNodeByPathAndAttr(Database, RECORDSTAG + '/' + BookTag, 'path', FN, Node);
end;

function TSonyPRS505.IsRegisteredBook(Ebook: TEbookFile): boolean;
var
  Dummy: TXmlNode;
begin
  Result := FindBookNode(Ebook, Dummy);
end;

function TSonyPRS505.CreateCollection(const Name, NameSort: string; Ebooks: TEbookFileList): boolean;
var
  ColNode, ItemNode: TXmlNode;
  i: integer;
  ID: string;
begin
  Connect;
  if XmlNodeByPath(Database, RECORDSTAG, ColNode) then
    begin
    ColNode := ColNode.NodeNew(UTF8String(PlaylistTag));
    ColNode.AttributeByNameWide['title'] := Name;
    ColNode.AttributeByNameWide['sourceid'] := SourceID;
    ColNode.AttributeByNameWide['id'] := IntToStr(NextID);
    ColNode.AttributeByNameWide['uuid'] := GenerateUUID;
    for i := 0 to Pred(Ebooks.Count) do
      if FindBookNode(Ebooks[i], ItemNode) then
        if ItemNode.HasAttribute('id') then
          begin
          ID := ItemNode.AttributeByNameWide['id'];
          ItemNode := ColNode.NodeNew(UTF8String(PlaylistItemTag));
          ItemNode.AttributeByNameWide['id'] := ID;
          end;
    Result := True;
    end
  else
    Result := False;
end;

function TSonyPRS505.DeleteAllCollections: boolean;
var
  Node: TXmlNode;
  i: integer;
begin
  Connect;
  if XmlNodeByPath(Database, RECORDSTAG, Node) then
    for i := Pred(Node.NodeCount) downto 0 do
      if Node.Nodes[i].Name = UTF8String(PlaylistTag) then
        Node.Nodes[i].Delete;
  Result := True;
end;

function TSonyPRS505.GetMimeType(Ebook: TEbookFile): string;
var
  Ext: string;
  i: integer;
begin
  Result := '';
  Ext := ExtractFileExt(Ebook.FileName);
  for i := 0 to Pred(Length(MimeTypes)) do
    if AnsiCompareText(Ext, MimeTypes[i].Extension) = 0 then
      begin
      Result := MimeTypes[i].MimeType;
      Break;
      end;
end;

function TSonyPRS505.NextId: integer;
begin
  Result := Succ(GetNextId(Database.Root));
end;

function TSonyPRS505.GetNextId(Node: TXmlNode): integer;
var
  i, NodeId: integer;
begin
  Result := 0;
  if Node <> nil then
    begin
    if Node.HasAttribute('id') then
      Result := StrToIntDef(Node.AttributeByNameWide['id'], Result);
    for i := 0 to Pred(Node.NodeCount) do
      begin
      NodeId := GetNextId(Node.Nodes[i]);
      if NodeId > Result then
        Result := NodeId;
      end;
    end;
end;

function TSonyPRS505.CreateBook(Ebook: TEbookFile; out Node: TXmlNode): boolean;
begin
  Connect;
  if XmlNodeByPath(Database, RECORDSTAG, Node) then
    begin
    Node := Node.NodeNew(UTF8String(BookTag));
    Node.AttributeByNameWide['author'] := Ebook.Author;
    //Node.AttributeByNameWide['authorSorter'] := Ebook.AuthorSort;
    Node.AttributeByNameWide['status'] := '0';
    Node.AttributeByNameWide['sourceid'] := SourceID;
    Node.AttributeByNameWide['id'] := IntToStr(NextId);
    Node.AttributeByNameWide['date'] := DateTime2Xml(BookDateTime(Ebook));
    Node.AttributeByNameWide['mime'] := GetMimeType(Ebook);
    Node.AttributeByNameWide['path'] := FindFilePathRelativeToRoot(Ebook);
    Node.AttributeByNameWide['size'] := IntToStr(Ebook.Size);
    Node.AttributeByNameWide['title'] := Ebook.Title;
    //Node.AttributeByNameWide['titleSorter'] := Ebook.TitleSort;
    Result := True;
    end
  else
    Result := False;
end;

function TSonyPRS505.UpdateBook(Ebook: TEbookFile; Node: TXmlNode): boolean;
begin
  Connect;
  Node.AttributeByNameWide['author'] := Ebook.Author;
  //Node.AttributeByNameWide['authorSorter'] := Ebook.AuthorSort;
  Node.AttributeByNameWide['date'] := DateTime2Xml(BookDateTime(Ebook));
  Node.AttributeByNameWide['size'] := IntToStr(Ebook.Size);
  Node.AttributeByNameWide['title'] := Ebook.Title;
  Node.AttributeByNameWide['titleSorter'] := Ebook.TitleSort;
  Result := True;
end;

function TSonyPRS505.UpdateBookThumbnail(Ebook: TEbookFile; Node: TXmlNode): boolean;
var
  Bitmap: TBitmap;
  Thumb: TJpegImage;
  ThumbStream: TMemoryStream;
  ThumbDataRaw: AnsiString;
  ThumbDataBase64: string;
  ThumbNode, ImageNode: TXmlNode;
begin
  Result := False;
  if Ebook.Cover <> nil then
    begin
    Bitmap := TBitmap.Create;
    try
      Bitmap.SetSize(THUMB_WIDTH, THUMB_HEIGHT);
      Bitmap.Canvas.StretchDraw(Rect(0, 0, Pred(Bitmap.Width), Pred(Bitmap.Height)), Ebook.Cover);
      Thumb := TJpegImage.Create;
      try
        Thumb.Assign(Bitmap);
        ThumbStream := TMemoryStream.Create;
        try
          Thumb.SaveToStream(ThumbStream);
          SetLength(ThumbDataRaw, ThumbStream.Size);
          Move(ThumbStream.Memory^, ThumbDataRaw[1], ThumbStream.Size);
          ThumbDataBase64 := string(EncodeBase64(ThumbDataRaw));
        finally
          FreeAndNil(ThumbStream);
          end;
        Connect;
        if not XmlNodeByPath(Node, ThumbnailTag, ThumbNode) then
          ThumbNode := Node.NodeNew(UTF8String(ThumbnailTag));
        ThumbNode.AttributeByNameWide['width'] := IntToStr(THUMB_WIDTH);
        ThumbNode.AttributeByNameWide['height'] := IntToStr(THUMB_HEIGHT);
        if not XmlNodeByPath(ThumbNode, ImageTag, ImageNode) then
          ImageNode := ThumbNode.NodeNew(UTF8String(ImageTag));
        ImageNode.ValueAsUnicodeString := ThumbDataBase64;
        Result := True;
      finally
        FreeAndNil(Thumb);
        end;
    finally
      FreeAndNil(Bitmap);
      end;
    end;
end;

function TSonyPRS505.RegisterBook(Ebook: TEbookFile): boolean;
var
  Node: TXmlNode;
begin
  if FindBookNode(Ebook, Node) then
    Result := UpdateBook(Ebook, Node)
  else
    Result := CreateBook(Ebook, Node);
  if Result then
    UpdateBookThumbnail(Ebook, Node);
end;

function TSonyPRS505.SupportedMediaTypes: TMediaTypes;
begin
  Result := [mtInternalMemory, mtSDcard, mtMScard];
end;

initialization
  TDevice.RegisterDeviceClass('prs-505', TSonyPRS505);

end.
